# a simple interface to turtle field for joho-kiso
# Yoshinori Hayakawa
# 10-JUL-2013
# 30-JUN-2017

require 'socket'

PORTNUMBER=2222
EVENT_HIT_BY_BULLET   = (0x00000001)
EVENT_RUN_INTO_TURTLE = (0x00000002)
EVENT_RUN_INTO_DONUT  = (0x00000004)
EVENT_RUN_INTO_STONE  = (0x00000008)
EVENT_RUN_INTO_WALL   = (0x00000010)
EVENT_FOUND_COIN      = (0x00000020)
EVENT_DETECTED_BY_FINDER = (0x00000040)
EVENT_DETECTED_BY_RADAR  = (0x00000080)
EVENT_GOT_MESSAGE        = (0x00000100)

class Turtle
  @socket 
  @last_message
  @team

  def Turtle.finalize(socket)
    proc { 
      if socket && !socket.closed?
        socket.close
      end
    }
  end

  def initialize(hostname="localhost")
    begin
      @socket = TCPSocket.open(hostname, PORTNUMBER)
    rescue
      printf("cannot connect to %s\n",hostname) 
      printf("seems like turtle field is not ready...\n") 
      exit
    end
    ObjectSpace.define_finalizer(self, Turtle.finalize(@socket))
    @team = 0 
  end

  def ttl_send_command(cmd)
    rc = 0 ;
    begin
      @socket.print(cmd + "\n")
      @last_message = @socket.gets
    rescue
      printf("I was killed.. bye\n") ;
      exit 
    else
      if @last_message.nil? or @last_message.length<=1 then
        printf("I was killed... bye\n") ;
        exit
      end
    end
    ack = @last_message.split(nil)
    eventcode = ack[-1].to_i
    timecode = ack[-2].to_i
    if ack[-4]=="OK" then
      rc = 0 
    else
      rc = -1
    end
    if eventcode != 0 then
      check_event(eventcode,timecode) 
    end
    sleep(0.001)
    return rc 
  end

  def enable_event(eventcode)
    ttl_send_command("Y + " + eventcode.to_s)
  end

  def disable_event(eventcode)
    ttl_send_command("Y - " + eventcode.to_s)
  end

  def check_event(event,timecode)
    msg_stack = @last_message.dup 
    myteam = @team
    if event & EVENT_HIT_BY_BULLET !=0 then
      disable_event(EVENT_HIT_BY_BULLET)
      hit_by_bullet(timecode)
      enable_event(EVENT_HIT_BY_BULLET)
    end
    if event & EVENT_RUN_INTO_TURTLE != 0 then
      disable_event(EVENT_RUN_INTO_TURTLE) 
      run_into_turtle(timecode)
      enable_event(EVENT_RUN_INTO_TURTLE) 
    end
    if event & EVENT_RUN_INTO_DONUT != 0 then
      disable_event(EVENT_RUN_INTO_DONUT) 
      run_into_donut(timecode)
      enable_event(EVENT_RUN_INTO_DONUT) 
    end
    if event & EVENT_RUN_INTO_STONE != 0 then
      disable_event(EVENT_RUN_INTO_STONE)
      run_into_stone(timecode)
      enable_event(EVENT_RUN_INTO_STONE)
    end
    if event & EVENT_RUN_INTO_WALL != 0 then
      disable_event(EVENT_RUN_INTO_WALL) 
      run_into_wall(timecode)
      enable_event(EVENT_RUN_INTO_WALL) 
    end
    if event & EVENT_FOUND_COIN != 0 then
      disable_event(EVENT_FOUND_COIN) 
      found_coin(timecode)
      enable_event(EVENT_FOUND_COIN)
    end
    if event & EVENT_DETECTED_BY_FINDER != 0 then
      disable_event(EVENT_DETECTED_BY_FINDER)
      detected_by_finder(timecode)
      enable_event(EVENT_DETECTED_BY_FINDER)
    end
    if event & EVENT_DETECTED_BY_RADAR != 0 then
      disable_event(EVENT_DETECTED_BY_RADAR)
      detected_by_radar(timecode)
      enable_event(EVENT_DETECTED_BY_RADAR)
    end
    if event & EVENT_GOT_MESSAGE != 0 then
      disable_event(EVENT_GOT_MESSAGE)
      got_message(timecode)
      enable_event(EVENT_GOT_MESSAGE) 
    end
    @last_message = msg_stack 
    @team = myteam 
  end

  def convert_str2hex(str)
    hex=""
    str.each_byte do |c| 
      hex << sprintf("%02X",c)
    end
    return hex
  end

  def char2int(c)
    r=0
    if c >= "0"[0] and c <= "9"[0] then
      r = c - "0"[0]
    elsif c>="a"[0] and c<="f"[0] then
      r = c - "a"[0] + 10 
    elsif c>="A"[0] and c<="F"[0] then
      r = c - "A"[0] + 10 
    end
    return r ;
  end

  def convert_hex2str(hex)
    str=""
    d0=0 
    d1=0
    cnt=0
    hex.each_byte do |c|
      next if c.chr == " "
      if cnt%2 == 0 then
        d1 = char2int(c)
      else
        d0 = char2int(c)
        str << "%c" % (d1*16+d0)
      end
      cnt=cnt+1
    end
    return str
  end

  def hit_by_bullet(timecode)
  end

  def run_into_turtle(timecode)
  end

  def run_into_donut(timecode)
  end

  def run_into_stone(timecode)
  end

  def run_into_wall(timecode)
  end

  def found_coin(timecode)
  end

  def detected_by_finder(timecode)
  end

  def detected_by_radar(timecode)
  end

  def got_message(timecode)
  end

  def home
    return ttl_send_command("C H") ;
  end

  def rst
    return ttl_send_command("C") 
  end

  def cf
    return ttl_send_command("C F") ;
  end

  def clr
    return ttl_send_command("C A") ;
  end

  def fd(step)
    return ttl_send_command("F " + step.to_s)
  end

  def bk(step)
    return ttl_send_command("B " + step.to_s) 
  end

  def jump(x,y)
    return ttl_send_command("J %f %f"%[x,y]) 
  end

  def rjump(dx,dy)
    return ttl_send_command("G %f %f"%[dx,dy]) 
  end

  def rt(angle)
    return ttl_send_command("T R "+angle.to_s) 
  end

  def lt(angle)
    return ttl_send_command("T L "+angle.to_s) 
  end

  def east()
    return ttl_send_command("T E") 
  end

  def north()
    return ttl_send_command("T N") 
  end

  def west()
    return ttl_send_command("T W") 
  end

  def south()
    return ttl_send_command("T S") 
  end

  def lw(width)
    return ttl_send_command("W " + width.to_s)
  end

  def pd
    return ttl_send_command("P D") 
  end

  def pu
    return ttl_send_command("P U") 
  end

  def fill
    return ttl_send_command("L +") ;
  end

  def brush
    return ttl_send_command("L B") ;
  end

  def line
    return ttl_send_command("L -") ;
  end

  def point
    return ttl_send_command("L P") ;
  end

  def idle(t)
    sleep t
  end

  def rf(range)
    return ttl_send_command("E "+range.to_s)
  end

  def say(something)    
    return ttl_send_command("S "+convert_str2hex(something)) 
  end

  def bcas(msg)    
    return ttl_send_command("U "+convert_str2hex(msg)) 
  end

  def tprintf(format, *varg)
    msg = format%varg
    return say(msg)
  end

  def tscanf(format)
    return @last_message.scanf(format)
  end

  def fire
    return ttl_send_command("X") 
  end

  def robot
    return ttl_send_command("R") 
  end

  def donut
    return ttl_send_command("D") 
  end
  alias doughnut donut

  def coin
    return ttl_send_command("O X") 
  end

  def borrowcoin
    return ttl_send_command("O B") 
  end
  alias borrow_coin borrowcoin

  def dropcoin
    return ttl_send_command("O S") 
  end
  alias drop_coin dropcoin

  def pickcoin
    return ttl_send_command("O R") 
  end
  alias pick_coin pickcoin

  def col(r, g, b)
    return ttl_send_command("K %f %f %f"%[r,g,b])
  end

  def bgc(r, g, b)
    return ttl_send_command("H %f %f %f"%[r,g,b])
  end

  def card(c, x, y)
    return ttl_send_command("Z %d %f %f"%[c,x,y])
  end

  def clrcard
    return ttl_send_command("Z 0")
  end

  def bmode
    return ttl_send_command("! +")
  end

  def gmode
    return ttl_send_command("! -")
  end

  def nm(name)
    return ttl_send_command("N "+convert_str2hex(name))
  end

  def tm(team)
    return ttl_send_command("I %d"%[team])
  end

  def q_id
    ttl_send_command("Q M")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_pos
    ttl_send_command("Q P")
    ack = @last_message.split(nil)
    return [ack[0].to_f, ack[1].to_f]
  end

  def q_dir
    ttl_send_command("Q D")
    ack = @last_message.split(nil)
    return ack[0].to_f
  end
  alias q_heading q_dir


  def q_alive
    begin
      @socket.print("Q A\n")
      @last_message = @socket.gets
    rescue
      return false
    else
      if not @last_message.nil? and @last_message.length>0 then
        return true
      else
        return false 
      end
    end
  end

  def q_score
    ttl_send_command("Q S")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_nt
    ttl_send_command("Q N")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_coin
    ttl_send_command("Q C")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_mycoin
    ttl_send_command("Q O")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_radar
    ttl_send_command("Q R")
    ack = @last_message.split(nil)
    @team = ack[1].to_i
    return ack[0].to_f
  end

  def q_radar_n(nn)
    ttl_send_command("Q R %d"%nn)
    ack = @last_message.split(nil)
    @team = ack[1].to_i
    return ack[0].to_f
  end

  def q_finder
    ttl_send_command("Q F")
    ack = @last_message.split(nil)
    @team = ack[1].to_i
    return ack[0].to_i
  end

  def q_tm
    return @team
  end

  def q_sonar
    ttl_send_command("Q W")
    ack = @last_message.split(nil)
    if ack[0].to_i > 0 then
      left=true
    else
      left=false
    end
    if ack[1].to_i > 0 then
      front=true
    else
      front=false
    end
    if ack[2].to_i > 0 then
      right=true
    else
      right=false
    end
    return [left, front, right]
  end

  def q_bcas
    ttl_send_command("Q B")
    ack = @last_message.split(nil)
    return [ack[0].to_i, convert_hex2str(ack[1])]
  end

  def q_time
    ttl_send_command("Q T")
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_stat_card(c)
    ttl_send_command("Q Z S %d"%c)
    ack = @last_message.split(nil)
    return ack[0].to_i
  end

  def q_pos_card(c)
    ttl_send_command("Q Z P %d"%c)
    ack = @last_message.split(nil)
    return  [ack[0].to_f, ack[1].to_f]
  end
end
